/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_module.h,v 1.101 2004/10/22 23:56:57 phargrov Exp $
 */

#ifndef _CR_MODULE_H
#define _CR_MODULE_H        1

#include <linux/list.h>
#include <asm/atomic.h>
#include <asm/rwlock.h>

#include <blcr_common.h>	// For the stuff common to user space

#include "blcr_config.h"	// Configuration
#include "cr_barrier.h"		// Code for kernel barriers.
#include "cr_ktrace.h"		// Code for trace messages

#ifdef __i386__
static inline struct pt_regs *
get_pt_regs(struct task_struct *tsk)
{
    return (struct pt_regs *)((unsigned long)(tsk) +
	                      sizeof(init_task_union.stack) -
		              sizeof(struct pt_regs));
}

static inline unsigned long
get_stack_pointer(struct pt_regs *regs)
{
    return regs->esp;
}
#endif

typedef struct cr_objectmap_s *cr_objectmap_t;

// Destination info for a checkpoint
typedef	struct cr_chkpt_dest_s {
	// file pointer (NULL iff destination is a directory)
	struct file *		filp;
	struct semaphore	mutex;

	// fs struct pointer (NULL iff destination is not a directory)
	struct fs_struct *	fs;
} cr_chkpt_dest_t;

typedef struct cr_rstrt_src_s {
        // file pointer (NULL iff destination is a directory)
        struct file *           filp;
        struct semaphore        mutex;

        // fs struct pointer (NULL iff destination is not a directory)
        struct fs_struct *      fs;
} cr_rstrt_src_t;


// Kernel-side tracking of a checkpoint request
//
// XXX: this needs a lot of stuff added as we progress:
// It will also keep such things as the count of tasks in the
// checkpoint, reference counts for resources, etc.
//
// XXX: the naming between user and kernel could be improved.
typedef struct cr_chkpt_req_s {
	atomic_t		ref_count;	// reference count
	atomic_t		completed;	// count of successful completions
	rwlock_t		lock;		// lock for task_list & wait
	wait_queue_head_t	wait;		// place for requester to wait
	struct list_head	tasks;		// list of target cr_task_t's
	cr_chkpt_dest_t		dest;		// checkpoint destination
	unsigned int		flags;		// flags supplied by requestor
	int			result;		// value returned by REAP
	int			has_expiration;	// 'expiration' is used
	unsigned long		expiration;	// limit on time to chkpt

	cr_objectmap_t          map;

	// barriers: XXX: could/should be done per thread group?
	cr_barrier_t		phase_barrier;		// between phases 1 & 2
	cr_barrier_t		predump_barrier;	// at start of dump
	cr_barrier_t		postdump_barrier;	// at end of dump

	struct {
	    struct tq_struct	    task;
	    struct timer_list	    timer;
	}			watchdog;	// Watchdog task and timer

        cr_format_t             dump_format;    // Format to use for this dump

	struct semaphore        serial_mutex;   // globally order steps in req

	cr_bool_t               wrote_header;   // bool: need to write?
#if HAVE_2_6_SIGNAL_STRUCT
	cr_bool_t               wrote_signal_struct;   // bool: need to write?
#endif
	cr_bool_t               wrote_linkage;   // bool: need to write?
	cr_bool_t               wrote_creds;	// bool: need to write?
	cr_bool_t               wrote_fs_struct;   // bool: need to write?
	cr_bool_t               wrote_files_struct;   // bool: need to write?
	cr_scope_t              checkpoint_scope;

	int			die;		// flag indicating we must die
	int			signal;		// sent after CHECKPOINT
} cr_chkpt_req_t;

typedef struct cr_rstrt_req_s {
        cr_rstrt_src_t          src;
        struct file            *file;
	cr_bool_t               need_linkage;
        cr_bool_t               need_fs;
        cr_bool_t               need_files;
	cr_bool_t               need_hide_cr_fds;
	cr_bool_t               need_close_cr_fds;
	cr_bool_t               need_children;
	cr_bool_t               need_clones;
	cr_bool_t               need_creds;
	atomic_t                new_clones;
	int                     total_threads;
	int                     shared_signal;
	atomic_t                ref_count;
	struct task_struct     *cr_restart_task;
	struct cr_context_tasklinkage *linkage;
	int                     linkage_size;
	struct semaphore        mutex;
	/* barrier for cr_restore_linkage */
	atomic_t                link_barrier_count;
	wait_queue_head_t       link_barrier_queue;
	/* place for requester to wait for us to finish */
	wait_queue_head_t	wait;
	int			result;		// value returned by REAP
	int			die;		// indicates spawned threads must die
	cr_objectmap_t          map;
} cr_rstrt_req_t;

// For forming a list of tasks.
typedef struct cr_task_s {
	struct list_head	task_list;
	struct list_head	req_list;
	atomic_t		ref_count;
	struct task_struct	*task;
	cr_chkpt_req_t		*req;
	struct file		*phase1_filp;
	struct file		*phase2_filp;
	void			*handler;	// XXX: type?
} cr_task_t;

// Private data attached to an instance of the file
typedef struct cr_pdata_s {
	cr_chkpt_req_t		*chkpt_req;
	// cr_chkpt_opt_t	*chkpt_opt;	NOT YET DEFINED
	cr_rstrt_req_t	        *rstrt_req;
} cr_pdata_t;

// Basic adaptation of timespec_to_jiffies from include/linux/time.h
static __inline__ unsigned long
timeval_to_jiffies(struct timeval *value)
{
	unsigned long sec = value->tv_sec;
	long usec = value->tv_usec;

	if (sec >= (MAX_JIFFY_OFFSET / HZ))
		return MAX_JIFFY_OFFSET;
	usec += 1000000L / HZ - 1;
	usec /= 1000000L / HZ;
	return HZ * sec + usec;
}

// Basic adaptation of jiffies_to_timeval from include/linux/time.h
static __inline__ void
jiffies_to_timeval(unsigned long jiffies, struct timeval *value)
{
	value->tv_usec = (jiffies % HZ) * (1000000L / HZ);
	value->tv_sec = jiffies / HZ;
}

#ifndef list_for_each_safe
// list_for_each_safe: A list iterator allowing for deletions
// First appears in 2.4.10 kernel, from which we've taken the following:
  #define list_for_each_safe(pos, n, head) \
    for (pos = (head)->next, n = pos->next;\
         pos != (head);\
         pos = n, n = pos->next)
#endif

// Process table iterators
// Designed to work for both new and old kernels.
//
// When using CR_FOR_EACH_TASK_VARS, either place it at the *end* of the
// declarations and/or don't place a semicolon after it (in case it's empty).
// The preferred option is to place it at the end of the declarations and
// use a semicolon.  This renders auto-indenting editors easier to live with.
#if HAVE_2_6_TASK_LINKAGE
  #define CR_FOR_EACH_TASK_VARS \
    struct list_head *_l; \
    struct pid *_p;
  #define CR_FOR_EACH_TASK_PGID(ID, T) \
    for_each_task_pid((ID), PIDTYPE_PGID, (T), _l, _p)
  #define CR_FOR_EACH_TASK_SID(ID, T) \
    for_each_task_pid((ID), PIDTYPE_SID, (T), _l, _p)
  #define CR_FOR_EACH_TASK_TGID(ID, T) \
    for_each_task_pid((ID), PIDTYPE_TGID, (T), _l, _p)
  #define CR_FOR_EACH_TASK_PROC(P, T) \
    if ((P)->mm) \
      CR_FOR_EACH_TASK_PGID((P)->pgrp, (T)) \
        if ((T)->mm == (P)->mm)
#elif HAVE_2_4_TASK_LINKAGE
  #define CR_FOR_EACH_TASK_VARS
  #define CR_FOR_EACH_TASK_PGID(ID, T) \
    for_each_task((T)) if ((T)->pgrp == (ID)) 
  #define CR_FOR_EACH_TASK_SID(ID, T) \
    for_each_task((T)) if ((T)->session == (ID)) 
  #define CR_FOR_EACH_TASK_PROC(P, T) \
    if ((P)->mm) \
      for_each_task((T)) \
        if ((T)->mm == (P)->mm) 
#else
  #error No tasklist iterators available
#endif

/* What is the opposite of "get_task_struct" ? */
#if HAVE_FREE_TASK_STRUCT
  #define put_task_struct free_task_struct
#elif HAVE_PUT_TASK_STRUCT
  #define __put_task_struct CR_IMPORT_KSYM(__put_task_struct)
#else
  #error
#endif

/* How do we manipulate the mmap_sem? */
#if HAVE_RW_MMAP_SEM
  #define cr_mmap_sem_down_read		down_read
  #define cr_mmap_sem_down_write	down_write
  #define cr_mmap_sem_up_read		up_read
  #define cr_mmap_sem_up_write		up_write
#else
  #define cr_mmap_sem_down_read		down
  #define cr_mmap_sem_down_write	down
  #define cr_mmap_sem_up_read		up
  #define cr_mmap_sem_up_write		up
#endif

/* Import non-exported kernel symbols */
#include "blcr_ksyms.h"
extern void * sys_call_table[];
#if CR_KSYM_sys_call_table
  #define sys_call_table CR_IMPORT_KSYM(sys_call_table)
#endif
#if CR_KSYM_do_fork
  #define do_fork CR_IMPORT_KSYM(do_fork)
#endif
#if CR_KSYM_detach_pid
  #define detach_pid CR_IMPORT_KSYM(detach_pid)
#endif
#if CR_KSYM_attach_pid
  #define attach_pid CR_IMPORT_KSYM(attach_pid)
#endif
#if CR_KSYM_link_pid
  #define link_pid CR_IMPORT_KSYM(link_pid)
#endif
#if CR_KSYM_find_pid
  #define find_pid CR_IMPORT_KSYM(find_pid)
#endif
/* symbols for pipes */
#if CR_KSYM_do_pipe
  #define do_pipe CR_IMPORT_KSYM(do_pipe)
#endif
#if CR_KSYM_write_pipe_fops
  #define write_pipe_fops CR_IMPORT_KSYM(write_pipe_fops)
#endif
#if CR_KSYM_rdwr_pipe_fops
  #define rdwr_pipe_fops CR_IMPORT_KSYM(rdwr_pipe_fops)
#endif
#if CR_KSYM_read_pipe_fops
  #define read_pipe_fops CR_IMPORT_KSYM(read_pipe_fops)
#endif
/* symbols for fifos */
#if CR_KSYM_write_fifo_fops
  extern struct file_operations write_fifo_fops;
  #define write_fifo_fops CR_IMPORT_KSYM(write_fifo_fops)
#endif
#if CR_KSYM_rdwr_fifo_fops
  extern struct file_operations rdwr_fifo_fops;
  #define rdwr_fifo_fops CR_IMPORT_KSYM(rdwr_fifo_fops)
#endif
#if CR_KSYM_read_fifo_fops
  extern struct file_operations read_fifo_fops;
  #define read_fifo_fops CR_IMPORT_KSYM(read_fifo_fops)
#endif
#if CR_KSYM_sys_setresuid
  extern long sys_setresuid(uid_t, uid_t, uid_t);
  #define sys_setresuid CR_IMPORT_KSYM(sys_setresuid)
#endif
#if CR_KSYM_sys_setresgid
  extern long sys_setresgid(gid_t, gid_t, gid_t);
  #define sys_setresgid CR_IMPORT_KSYM(sys_setresgid)
#endif
#if CR_KSYM_sys_setgroups
  extern long sys_setgroups(int, gid_t *);
  #define sys_setgroups CR_IMPORT_KSYM(sys_setgroups)
#endif

// cr_proc.c
extern int cr_proc_init(void);
extern void cr_proc_cleanup(void);

// cr_fops.c
extern struct file_operations cr_ctrl_fops;
extern struct file_operations cr_generic_bogus_fops;
extern int cr_fs_init(void);
extern int cr_fs_cleanup(void);

// cr_dump_self.c
extern int cr_dump_self(struct file *filp, unsigned long flags);
extern int cr_cont_complete(struct file *filp);
extern char *cr_getpath(struct dentry *dentry, struct vfsmount *mnt,
			char *buf, int size);

// cr_chkpt_req.c
extern int cr_chkpt_req(struct file *filp, void *arg);
extern int cr_chkpt_done(struct file *filp, void *arg);
extern int cr_chkpt_reap(struct file *filp);
extern void cr_chkpt_task_complete(cr_task_t *cr_task);
extern void cr_chkpt_req_release(struct file *filp, cr_pdata_t *priv);
extern int cr_abort_checkpoint(struct file *filp, unsigned long flags);
extern int cr_chkpt_dest(struct file *filp, char *arg);

// cr_async.c
extern int cr_suspend(struct file *filp, void *arg);
extern int cr_phase1_register(struct file *filp, void *arg);
extern void cr_phase1_release(struct file *filp, cr_pdata_t *priv);

// cr_sync.c
extern int cr_phase2_register(struct file *filp, void *arg);
extern void cr_phase2_release(struct file *filp, cr_pdata_t *priv);

// cr_task.c
extern rwlock_t cr_task_lock;
extern struct list_head cr_task_list;
extern cr_task_t *__cr_task_get(struct task_struct *task, int create);
extern void __cr_task_put(cr_task_t *cr_task);
extern cr_task_t *cr_task_get(struct task_struct *task);
extern void cr_task_put(cr_task_t *cr_task);

// cr_rstrt_req.c
extern int cr_rstrt_request_restart(struct file *filp, void *arg);
extern int cr_rstrt_done(struct file *filp, void *arg);
extern int cr_rstrt_reap(struct file *filp);
extern int cr_rstrt_child(struct file *filp);
extern int cr_rstrt_barrier(struct file *filp, unsigned long flags);
extern int cr_rstrt_req_release(struct file *filp, cr_pdata_t *priv);

// cr_dest_file.c
extern int cr_dest_init(cr_chkpt_dest_t *dest, int fd, struct file *from);
extern void cr_dest_free(cr_chkpt_dest_t *dest);
extern struct file *cr_dest_get(cr_chkpt_dest_t *dest, int *shared);
extern void cr_dest_put(cr_chkpt_dest_t *dest, struct file *filp);

// cr_trigger.c
extern int cr_trigger_phase1(cr_chkpt_req_t *req);
extern int cr_trigger_phase2(cr_chkpt_req_t *req);

// cr_io.c
extern ssize_t cr_kread(struct file * file, void *buf, size_t count);
extern ssize_t cr_kwrite(struct file * file, void *buf, size_t count);

// cr_objects.c
extern cr_objectmap_t cr_alloc_objectmap(void);
extern void cr_release_objectmap(cr_objectmap_t);
int cr_find_object(cr_objectmap_t, void *, void **);
int cr_insert_object(cr_objectmap_t, void *, void *);
int cr_remove_object(cr_objectmap_t, void *);

#endif
