/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_objects.c,v 1.1 2004/02/12 00:10:38 eroman Exp $
 */

#include <linux/module.h>
#include <linux/mm.h>

#include "cr_module.h"

#define MAP_SIZE_BYTES PAGE_SIZE

struct cr_objectmap_s {
    void *id;
    void *ptr; 
};

rwlock_t objectmap_lock = RW_LOCK_UNLOCKED;

static inline int
mapsize(cr_objectmap_t map)
{
    return ((MAP_SIZE_BYTES)/(sizeof(struct cr_objectmap_s)));
}

cr_objectmap_t
cr_alloc_objectmap(void)
{
    struct cr_objectmap_s *map;
    int i;

    /* should make this bigger at some point */
    map = (struct cr_objectmap_s *) __get_free_page(GFP_KERNEL);
    if (map) {
        write_lock(&objectmap_lock);
        for (i=0; i<mapsize(map); ++i) {
            map[i].id = NULL;
	    map[i].ptr = NULL;
        }
        write_unlock(&objectmap_lock);
    }

    return ((cr_objectmap_t) map);
}

void
cr_release_objectmap(cr_objectmap_t map)
{
    free_page((unsigned long)map);
}

/*
 * returns int rather than void * to allow NULL id's to be placed into table
 *
 * 1 - found
 * 0 - not found
 */
int
cr_find_object(cr_objectmap_t map, void *id, void **ptr)
{
    int i;
    int retval;

    retval = 0;

    /* return NULL by default, use return value to distinguish good NULL from evil NULL */
    if (ptr != NULL) 
	  *ptr = NULL;

    /* special cased to avoid confusion with empty */
    if (id == NULL) {
        // CR_KTRACE_LOW_LVL("Asked for NULL, returning NULL.");
        retval = 1;
        return retval;
    }

    read_lock(&objectmap_lock);
    for (i=0; i<mapsize(map) && retval != 1; ++i) {
        if(map[i].id == id) {
            // CR_KTRACE_LOW_LVL("Found object %p in slot %d", id, i);
	    if (ptr != NULL)
	      *ptr = map[i].ptr;
            retval = 1; 
	}
    }
    read_unlock(&objectmap_lock);

    return retval;
}

/*
 *  1 if it's in there already 
 *  0 if we insert it
 */
int
cr_insert_object(cr_objectmap_t map, void *id, void *ptr)
{
    int retval = -1;
    int i;

    if (id == NULL) {
        retval = 1;
	return retval;
    }

    write_lock(&objectmap_lock);
    for (i=0; i<mapsize(map) && retval != 0; ++i) {
       if(map[i].id == id) {
             /* return 1 if it's in there already */
	     // CR_KTRACE_LOW_LVL("Object %p already inserted into slot %d", id, i);
             retval = 1;
	     goto out_unlock;
       }
    }

    for (i=0; i<mapsize(map) && retval != 0; ++i) {
        if (map[i].id == NULL) {
	    // CR_KTRACE_LOW_LVL("Inserting object %p into slot %d", id, i);
            map[i].id = id;
	    map[i].ptr = ptr;
	    retval = 0;
	}
    }

out_unlock:
    write_unlock(&objectmap_lock);
    return retval;
}

int
cr_remove_object(cr_objectmap_t map, void *id)
{
    int i;
    int retval=-1;

    write_lock(&objectmap_lock);
    for (i=0; i<mapsize(map); ++i) {
        if(map[i].id == id) {
	    map[i].id = NULL;
            retval = 0; 
	}
    }
    write_unlock(&objectmap_lock);

    return retval;
}
